import 'package:click_extensions/click_extensions.dart';
import 'package:click_nfse/src/exceptions/validator_exception.dart';
import 'package:click_nfse/src/request_mapping.dart';

class DocumentoModel extends RequestMapping {
  late final ParametrizacaoModel? parametrizacao;
  late final NFSeModel? nfse;
  DocumentoModel({
    required this.parametrizacao,
    required this.nfse,
  }) : super.empty();
  DocumentoModel.empty() : super.empty();
  DocumentoModel.fromJson(super.dataRequest) : super.fromJson();
  DocumentoModel.fromMap(super.map) : super.fromMap();

  @override
  void fromMap() {
    parametrizacao = ParametrizacaoModel.fromMap(data['parametrizacao']);
    nfse = NFSeModel.fromMap(data['nfse']);
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'parametrizacao': parametrizacao!.toMap(),
      'nfse': nfse!.toMap(),
    };
  }

  @override
  void validate() {
    if (parametrizacao == null) throw ValidatorException('Os campos de parametrização do documento deve ser informado');
    if (nfse == null) throw ValidatorException('Os campos da NFSe do documento deve ser informado');

    nfse!.servico.discriminacao =
        nfse!.itens.map((item) => '${item.quantidade!.toInt()} ${item.descricao} - ${item.valorTotal!.toFormattedMoney}').join(' | ');

    parametrizacao!.validate();
    nfse!.validate();
  }
}

class ParametrizacaoModel extends RequestMapping {
  late final bool? sandbox;
  late final int? numeroLote;
  late final int? layoutProvedor;
  late final int? codigoMunicipio;
  late final String? nomeMunicipio;
  late final String? ufMunicipio;
  late final String? usuario;
  late final String? senha;
  late final String? cnpjEmitente;
  late final String? imEmitente;
  late final String? nomeEmitente;

  ParametrizacaoModel.empty() : super.empty();
  ParametrizacaoModel.fromJson(super.dataRequest) : super.fromJson();
  ParametrizacaoModel.fromMap(super.map) : super.fromMap();
  ParametrizacaoModel({
    this.numeroLote,
    this.sandbox,
    this.layoutProvedor,
    this.codigoMunicipio,
    this.nomeMunicipio,
    this.ufMunicipio,
    this.usuario,
    this.senha,
    this.cnpjEmitente,
    this.imEmitente,
    this.nomeEmitente,
  }) : super.empty();

  @override
  void fromMap() {
    numeroLote = data['numeroLote'];
    sandbox = data['sandbox'] ?? true;
    layoutProvedor = data['layoutProvedor'];
    codigoMunicipio = data['codigoMunicipio'];
    nomeMunicipio = data['nomeMunicipio'];
    ufMunicipio = data['ufMunicipio'];
    usuario = data['usuario'];
    senha = data['senha'];
    cnpjEmitente = data['cnpjEmitente'];
    imEmitente = data['imEmitente'];
    nomeEmitente = data['nomeEmitente'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'numeroLote': numeroLote,
      'sandbox': sandbox,
      'layoutProvedor': layoutProvedor,
      'codigoMunicipio': codigoMunicipio,
      'nomeMunicipio': nomeMunicipio,
      'ufMunicipio': ufMunicipio,
      'usuario': usuario,
      'senha': senha,
      'cnpjEmitente': cnpjEmitente,
      'imEmitente': imEmitente,
      'nomeEmitente': nomeEmitente,
    };
  }

  @override
  void validate() {
    if (numeroLote == null || numeroLote! == 0) throw ValidatorException('Número do lote deve ser informado');
    if (sandbox! == false) {
      if (layoutProvedor == null) throw ValidatorException('O parametro layoutProvedor não foi informado');
      if (codigoMunicipio == null || codigoMunicipio! == 0) throw ValidatorException('O parametro codigoMunicipio não foi informado');
      if (nomeMunicipio == null || nomeMunicipio!.isEmpty) throw ValidatorException('O parametro nomeMunicipio não foi informado');
      if (ufMunicipio == null || ufMunicipio!.isEmpty) throw ValidatorException('O parametro ufMunicipio não foi informado');
      if (usuario == null || usuario!.isEmpty) throw ValidatorException('O parametro usuario não foi informado');
      if (senha == null || senha!.isEmpty) throw ValidatorException('O parametro senha não foi informado');
      if (cnpjEmitente == null || cnpjEmitente!.isEmpty) throw ValidatorException('O parametro cnpjEmitente não foi informado');
      if (imEmitente == null || imEmitente!.isEmpty) throw ValidatorException('O parametro imEmitente não foi informado');
      if (nomeEmitente == null || nomeEmitente!.isEmpty) throw ValidatorException('O parametro nomeEmitente não foi informado');
    }
  }
}

class NFSeModel extends RequestMapping {
  late final IdentificacaoRps identificacaoRps;
  late final IdentificacaoNFSe? identificacaoNFSe;
  late final RpsSubstituido? rpsSubstituido;
  late final Prestador prestador;
  late final Tomador tomador;
  late final Intermediario? intermediario;
  //late final ConstrucaoCivil? construcaoCivil;
  late final Servico servico;
  late final List<Itens> itens;
  late final Valores valores;
  //late final CondicaoPagamento condicaoPagamento;

  NFSeModel.empty() : super.empty();
  NFSeModel.fromJson(super.dataRequest) : super.fromJson();
  NFSeModel.fromMap(super.map) : super.fromMap();

  NFSeModel({
    required this.identificacaoRps,
    this.identificacaoNFSe,
    this.rpsSubstituido,
    required this.prestador,
    required this.tomador,
    this.intermediario,
    //this.construcaoCivil,
    required this.servico,
    required this.itens,
    //required this.valores,
    //required this.condicaoPagamento,
  }) : super.empty();

  @override
  void fromMap() {
    data['identificacaoRps'] != null
        ? identificacaoRps = IdentificacaoRps.fromMap(data['identificacaoRps'])
        : throw ValidatorException('Bloco identificacaoRps não foi informado');

    data['identificacaoNFSe'] != null ? identificacaoNFSe = IdentificacaoNFSe.fromMap(data['identificacaoNFSe']) : identificacaoNFSe = null;
    data['rpsSubstituido'] != null ? rpsSubstituido = RpsSubstituido.fromMap(data['rpsSubstituido']) : rpsSubstituido = null;

    data['prestador'] != null ? prestador = Prestador.fromMap(data['prestador']) : throw ValidatorException('Bloco prestador não foi informado');
    data['tomador'] != null ? tomador = Tomador.fromMap(data['tomador']) : throw ValidatorException('Bloco tomador não foi informado');

    data['intermediario'] != null ? throw ValidatorException('Bloco de intermediario não foi implementado') : intermediario = null;

    data['servico'] != null ? servico = Servico.fromMap(data['servico']) : throw ValidatorException('Bloco servico não foi informado');

    data['itens'] != null
        ? itens = (data['itens'] as List).map((itm) => Itens.fromMap(itm)).toList()
        : throw ValidatorException('Bloco Itens não foi informado');

    valores = Valores(
      valorServicos: itens.fold(0, (previousValue, element) => previousValue! + element.valorTotal!),
      valorDeducoes: itens.fold(0, (previousValue, element) => previousValue! + element.valorDeducoes!),
      valorPis: null,
      aliquotaPis: null,
      valorCofins: null,
      aliquotaCofins: null,
      valorInss: null,
      valorIr: null,
      valorCsll: null,
      iSSRetido: null,
      outrasRetencoes: null,
      descontoIncondicionado: itens.fold(0, (previousValue, element) => previousValue! + element.descontoIncondicionado!),
      descontoCondicionado: null,
      baseCalculo: itens.fold(0, (previousValue, element) => previousValue! + element.baseCalculo!),
      aliquota: itens[0].aliquota,
      valorIss: itens.fold(0, (previousValue, element) => previousValue! + element.valorIss!),
      valorIssRetido: 0,
      valorLiquidoNfse: itens.fold(0, (previousValue, element) => previousValue! + element.valorTotal!),
    );

    //   intermediario = data['intermediario'];
    //   construcaoCivil = data['construcaoCivil'];
    //   condicaoPagamento = data['condicaoPagamento'];
    // validate();
  }

  @override
  String toIni([int? itemNumber]) {
    String result = '';
    result += identificacaoRps.toIni();
    result += rpsSubstituido?.toIni() ?? '';
    result += prestador.toIni();
    result += tomador.toIni();
    result += servico.toIni();

    for (var item in itens) {
      result += item.toIni(itens.indexOf(item) + 1);
    }

    result += valores.toIni();
    return result;
  }

  @override
  Map<String, dynamic> toMap() {
    final result = <String, dynamic>{};

    result.addAll({'identificacaoRps': identificacaoRps.toMap()});
    if (identificacaoNFSe != null) result.addAll({'identificacaoNFSe': identificacaoNFSe!.toMap()});
    if (rpsSubstituido != null) result.addAll({'rpsSubstituido': rpsSubstituido!.toMap()});
    result.addAll({'prestador': prestador.toMap()});
    result.addAll({'tomador': tomador.toMap()});
    result.addAll({'servico': servico.toMap()});
    result.addAll({'itens': itens.map((e) => e.toMap()).toList()});
    result.addAll({'valores': valores.toMap()});

    //'intermediario': intermediario,
    //'construcaoCivil': construcaoCivil,
    //'itens': itens,
    //'condicaoPagamento': condicaoPagamento,

    return result;
  }

  @override
  void validate() {
    identificacaoRps.validate();
    if (identificacaoNFSe != null) identificacaoNFSe!.validate();
    if (rpsSubstituido != null) rpsSubstituido!.validate();
    prestador.validate();
    tomador.validate();
    servico.validate();
    for (var item in itens) {
      item.validate();
    }
  }
}

class IdentificacaoNFSe extends RequestMapping {
  late final int? numero;
  IdentificacaoNFSe({
    this.numero,
  }) : super.empty();
  IdentificacaoNFSe.empty() : super.empty();
  IdentificacaoNFSe.fromJson(super.dataRequest) : super.fromJson();
  IdentificacaoNFSe.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    numero = data['numero'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {'numero': numero};
  }

  @override
  void validate() {
    if (numero == 0) throw ValidatorException('Número não foi informado ou é inválido');
  }
}

class IdentificacaoRps extends RequestMapping {
  late final int numero;
  late final DateTime? competencia;
  late final String? situacaoTrib;
  late final int? producao;
  late final int? status;
  late final String? outrasInformacoes;
  IdentificacaoRps({
    required numero,
    required competencia,
    this.situacaoTrib,
    this.producao,
    this.status,
    this.outrasInformacoes,
  }) : super.empty();
  IdentificacaoRps.empty() : super.empty();
  IdentificacaoRps.fromJson(super.dataRequest) : super.fromJson();
  IdentificacaoRps.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    numero = data['numero'];
    competencia = data['competencia']?.toString().toDate;
    situacaoTrib = data['situacaoTrib'];
    producao = data['producao'];
    status = data['status'];
    outrasInformacoes = data['outrasInformacoes'];
  }

  @override
  Map<String, dynamic> toMap() {
    final map = <String, dynamic>{
      'numero': numero,
      'competencia': competencia!.toDateFullIntl,
      if (situacaoTrib != null) 'situacaoTrib': situacaoTrib,
      if (producao != null) 'producao': producao,
      if (status != null) 'status': status,
      if (outrasInformacoes != null) 'outrasInformacoes': outrasInformacoes,
    };

    return map;
  }

  @override
  void validate() {
    if (numero == 0) throw ValidatorException('Número do RPS não foi informado');
    if (competencia == null) throw ValidatorException('Competencia do RPS não foi informada');
    if (competencia!.isAfter(DateTime.now())) throw ValidatorException('Competencia do RPS informada é posterior a data atual');

    if (DateTime.now().add(Duration(days: -30)).isAfter(competencia!)) throw ValidatorException('Competencia do RPS informada é superior a 30 dias.');
  }
}

class RpsSubstituido extends RequestMapping {
  late final int? numero;
  late final int? serie;
  late final int? tipo;
  RpsSubstituido({
    this.numero,
    this.serie,
    this.tipo,
  }) : super.empty();

  RpsSubstituido.empty() : super.empty();
  RpsSubstituido.fromJson(super.dataRequest) : super.fromJson();
  RpsSubstituido.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    numero = data['numero'];
    serie = data['serie'];
    tipo = data['tipo'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      if (numero != null) 'numero': numero,
      if (serie != null) 'serie': serie,
      if (tipo != null) 'tipo': tipo,
    };
  }

  @override
  void validate() {
    if (numero == 0) throw ValidatorException('Número do RPS Substituido não foi informado');
    if (serie == 0) throw ValidatorException('Série do RPS Substituido não foi informado');
    if (tipo == 0) throw ValidatorException('Tipo do RPS Substituido não foi informado');
  }
}

class Prestador extends RequestMapping {
  ///1 - retMicroempresaMunicipal
  ///2 - retEstimativa
  ///3 - retSociedadeProfissionais
  ///4 - retCooperativa
  ///5 - retMicroempresarioIndividual
  ///6 - retMicroempresarioEmpresaPP
  ///7 - retLucroReal
  ///8 - retLucroPresumido
  ///9 - retSimplesNacional
  ///10- retImune
  ///11- retEmpresaIndividualRELI
  ///12- retEmpresaPP
  ///13- retMicroEmpresario
  ///14- retOutros
  late final int? regime;

  ///1 - Sim ou 2 - Não
  late final int? optanteSN;

  ///1 - Sim ou 2 - Não
  late final int? incentivadorCultural;
  late final String? cnpj;
  late final String? inscricaoMunicipal;
  late final String? razaoSocial; // Específico para provedor ISSDigital
  late final String? nomeFantasia; // Específico para provedor ISSDigital
  late final String? logradouro; // Específico para provedor ISSDigital
  late final int? numeroEndereco; // Específico para provedor ISSDigital
  late final String? bairro; // Específico para provedor ISSDigital
  late final String? codigoMunicipio; // Específico para provedor ISSDigital
  late final String? uf; // Específico para provedor ISSDigital
  late final int? codigoPais; // Específico para provedor ISSDigital
  late final String? xPais; // Específico para provedor ISSDigital
  late final String? cep; // Específico para provedor ISSDigital
  late final String? telefone; // Específico para provedor ISSDigital
  late final String? email; // Específico para provedor ISSDigital
  Prestador({
    this.regime,
    required this.optanteSN,
    required this.incentivadorCultural,
    required this.cnpj,
    required this.inscricaoMunicipal,
    this.razaoSocial,
    this.nomeFantasia,
    this.logradouro,
    this.numeroEndereco,
    this.bairro,
    this.codigoMunicipio,
    this.uf,
    this.codigoPais,
    this.xPais,
    this.cep,
    this.telefone,
    this.email,
  }) : super.empty();
  Prestador.empty() : super.empty();
  Prestador.fromJson(super.dataRequest) : super.fromJson();
  Prestador.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    regime = data['regime'];
    optanteSN = data['optanteSN'];
    incentivadorCultural = data['incentivadorCultural'];
    cnpj = data['cnpj'];
    inscricaoMunicipal = data['inscricaoMunicipal'];
    razaoSocial = data['razaoSocial']; // Específico para provedor ISSDigital
    nomeFantasia = data['nomeFantasia']; // Específico para provedor ISSDigital
    logradouro = data['logradouro']; // Específico para provedor ISSDigital
    numeroEndereco = data['numeroEndereco']; // Específico para provedor ISSDigital
    bairro = data['bairro']; // Específico para provedor ISSDigital
    codigoMunicipio = data['codigoMunicipio']; // Específico para provedor ISSDigital
    uf = data['uf']; // Específico para provedor ISSDigital
    codigoPais = data['codigoPais']; // Específico para provedor ISSDigital
    xPais = data['xPais']; // Específico para provedor ISSDigital
    cep = data['cep']; // Específico para provedor ISSDigital
    telefone = data['telefone']; // Específico para provedor ISSDigital
    email = data['email']; // Específico para provedor ISSDigital
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      if (regime != null) 'regime': regime,
      'optanteSN': optanteSN,
      'incentivadorCultural': incentivadorCultural,
      'cnpj': cnpj,
      'inscricaoMunicipal': inscricaoMunicipal,
      if (razaoSocial != null) 'razaoSocial': razaoSocial,
      if (nomeFantasia != null) 'nomeFantasia': nomeFantasia,
      if (logradouro != null) 'logradouro': logradouro,
      if (numeroEndereco != null) 'numeroEndereco': numeroEndereco,
      if (bairro != null) 'bairro': bairro,
      if (codigoMunicipio != null) 'codigoMunicipio': codigoMunicipio,
      if (uf != null) 'uf': uf,
      if (codigoPais != null) 'codigoPais': codigoPais,
      if (xPais != null) 'xPais': xPais,
      if (cep != null) 'cep': cep,
      if (telefone != null) 'telefone': telefone,
      if (email != null) 'email': email,
    };
  }

  @override
  void validate() {
    //if (regime == null || regime! <= 0) throw ValidatorException('Regime do prestador não foi informado');
    //if (regime! > 14) {
    //  throw ValidatorException(
    //      'Regime do prestador inválido: Valor experado: 1 - retMicroempresaMunicipal\n2 - retEstimativa\n3 - retSociedadeProfissionais \n4 - retCooperativa\n5 - retMicroempresarioIndividual\n6 - retMicroempresarioEmpresaPP\n7 - retLucroReal \n8 - retLucroPresumido \n9 - retSimplesNacional\n10- retImune \n11- retEmpresaIndividualRELI \n12- retEmpresaPP\n13- retMicroEmpresario  \n14- retOutros\n');
    //}

    if (optanteSN == null || optanteSN! <= 0) throw ValidatorException('OptanteSN do prestador não informado');
    if (optanteSN! > 2) throw ValidatorException('OptanteSN do prestador inválido: Valor experado: 1 - Sim ou 2 - Não');

    if (incentivadorCultural == null || incentivadorCultural! <= 0) throw ValidatorException('IncentivadorCultural do prestador não informado');
    if (incentivadorCultural! > 2) throw ValidatorException('IncentivadorCultural do prestador inválido: Valor experado: 1 - Sim ou 2 - Não');

    if (cnpj == null) throw ValidatorException('CNPJ do prestador não informado');
    if (!cnpj!.isValidCNPJ) throw ValidatorException('CNPJ do prestador é inválido, certifique-se que o cnpj foi digitado corretamente');

    if (inscricaoMunicipal == null || inscricaoMunicipal!.isEmpty) throw ValidatorException('InscricaoMunicipal do prestador não informado');
  }
}

class Tomador extends RequestMapping {
  late final int? tipo;
  late final String? cnpjCpf;
  late final String? inscricaoMunicipal;
  late final String? inscricaoEstadual;
  late final String? razaoSocial; // Específico para provedor ISSDigital
  late final String? tipoLogradouro; // Específico para provedor ISSDigital
  late final String? logradouro; // Específico para provedor ISSDigital
  late final int? numero; // Específico para provedor ISSDigital
  late final String? complemento; // Específico para provedor ISSDigital
  late final String? bairro; // Específico para provedor ISSDigital
  late final int? codigoMunicipio; // Específico para provedor ISSDigital
  late final String? uf; // Específico para provedor ISSDigital
  late final int? codigoPais; // Específico para provedor ISSDigital
  late final String? xPais; // Específico para provedor ISSDigital
  late final String? cep; // Específico para provedor ISSDigital
  late final String? telefone; // Específico para provedor ISSDigital
  late final String? email; // Específico para provedor ISSDigital
  late final int? atualizaTomador; // Específico para provedor ISSDigital
  late final int? tomadorExterior; // Específico para provedor ISSDigital
  Tomador({
    required this.tipo,
    required this.cnpjCpf,
    required this.inscricaoMunicipal,
    required this.inscricaoEstadual,
    required this.razaoSocial,
    required this.tipoLogradouro,
    required this.logradouro,
    required this.numero,
    this.complemento,
    required this.bairro,
    required this.codigoMunicipio,
    this.uf,
    this.codigoPais,
    this.xPais,
    required this.cep,
    this.telefone,
    this.email,
    this.atualizaTomador,
    this.tomadorExterior,
  }) : super.empty();

  Tomador.empty() : super.empty();
  Tomador.fromJson(super.dataRequest) : super.fromJson();
  Tomador.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    ///1:tpPFNaoIdentificada 2:tpPF 3:tpPJdoMunicipio 4:tpPJforaMunicipio 5:tpPJforaPais
    tipo = data['tipo'];
    cnpjCpf = data['cnpjCpf'];
    inscricaoMunicipal = data['inscricaoMunicipal'];
    inscricaoEstadual = data['inscricaoEstadual'];
    razaoSocial = data['razaoSocial']; // Específico para provedor ISSDigital
    tipoLogradouro = data['tipoLogradouro']; // Específico para provedor ISSDigital
    logradouro = data['logradouro']; // Específico para provedor ISSDigital
    numero = data['numero']; // Específico para provedor ISSDigital
    complemento = data['complemento']; // Específico para provedor ISSDigital
    bairro = data['bairro']; // Específico para provedor ISSDigital
    codigoMunicipio = data['codigoMunicipio']; // Específico para provedor ISSDigital
    uf = data['uf']; // Específico para provedor ISSDigital
    codigoPais = data['codigoPais']; // Específico para provedor ISSDigital
    xPais = data['xPais']; // Específico para provedor ISSDigital
    cep = data['cep']; // Específico para provedor ISSDigital
    telefone = data['telefone']; // Específico para provedor ISSDigital
    email = data['email']?.toString().toLowerCase(); // Específico para provedor ISSDigital
    atualizaTomador = data['atualizaTomador']; // Específico para provedor ISSDigital
    tomadorExterior = data['tomadorExterior']; // Específico para provedor ISSDigital
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'tipo': tipo,
      'cnpjCpf': cnpjCpf,
      'inscricaoMunicipal': inscricaoMunicipal,
      'inscricaoEstadual': inscricaoEstadual,
      'razaoSocial': razaoSocial,
      'tipoLogradouro': tipoLogradouro,
      'logradouro': logradouro,
      'numero': numero,
      if (complemento != null) 'complemento': complemento,
      'bairro': bairro,
      'codigoMunicipio': codigoMunicipio,
      if (uf != null) 'uf': uf,
      if (codigoPais != null) 'codigoPais': codigoPais,
      if (xPais != null) 'xPais': xPais,
      'cep': cep,
      if (telefone != null) 'telefone': telefone,
      if (email != null) 'email': email?.toLowerCase(),
      if (atualizaTomador != null) 'atualizaTomador': atualizaTomador,
      if (tomadorExterior != null) 'tomadorExterior': tomadorExterior,
    };
  }

  @override
  void validate() {
    if (tipo == null || tipo! <= 0) throw ValidatorException('Tipo da Seção [tipo] Tomador não informado');
    if (tipo! > 5) {
      throw ValidatorException(
          'Tipo da Seção do Tomador [tipo] inválido, Valores permitidos: 1 - tpPFNaoIdentificada\n2 - tpPF\n3 - tpPJdoMunicipio\n4 - tpPJforaMunicipio\n5 - tpPJforaPais\n');
    }
    if (cnpjCpf == null) throw ValidatorException('CNPJ ou CPF [cnpjCpf] do prestador não informado');
    if (!cnpjCpf!.isValidCPForCNPJ) {
      throw ValidatorException('CNPJ ou CPF [cnpjCpf] do prestador é inválido, certifique-se que o campo foi digitado corretamente');
    }
    if (inscricaoEstadual == null || inscricaoEstadual!.isEmpty) {
      throw ValidatorException('Inscrição Estadual [inscricaoEstadual] do tomador não informado');
    }
    if (razaoSocial == null || razaoSocial!.isEmpty) throw ValidatorException('Razão Social [razaoSocial] do tomador não informado');
    if (logradouro == null || logradouro!.isEmpty) throw ValidatorException('Logradouro [logradouro] do tomador não informado');
    if (numero == null) throw ValidatorException('Número [numero] do tomador não informado');
    if (bairro == null || bairro!.isEmpty) throw ValidatorException('Bairro [bairro] do tomador não informado');
    if (codigoMunicipio == null || codigoMunicipio! == 0) throw ValidatorException('Código do município [codigoMunicipio] do tomador não informado');
    if (cep == null || cep!.isEmpty) throw ValidatorException('Cep [cep] do tomador não informado');
    if (cep!.length != 8) throw ValidatorException('Cep [cep] do tomador deve conter 8 caracteres');
  }
}

class Intermediario extends RequestMapping {
  late final String? cnpjCpf;
  late final String? inscricaoMunicipal;
  late final String? razaoSocial;
  Intermediario({
    this.cnpjCpf,
    this.inscricaoMunicipal,
    this.razaoSocial,
  }) : super.empty();

  Intermediario.empty() : super.empty();
  Intermediario.fromJson(super.dataRequest) : super.fromJson();
  Intermediario.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    cnpjCpf = data['cnpjCpf'];
    inscricaoMunicipal = data['inscricaoMunicipal'];
    razaoSocial = data['razaoSocial'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'cnpjCpf': cnpjCpf,
      'inscricaoMunicipal': inscricaoMunicipal,
      'razaoSocial': razaoSocial,
    };
  }

  @override
  void validate() {
    // TODO: implement validate
  }
}

class ConstrucaoCivil extends RequestMapping {
  late final String? codigoObra;
  late final String? art;
  ConstrucaoCivil({
    this.codigoObra,
    this.art,
  }) : super.empty();
  ConstrucaoCivil.empty() : super.empty();
  ConstrucaoCivil.fromJson(super.dataRequest) : super.fromJson();

  @override
  void fromMap() {
    codigoObra = data['codigoObra'];
    art = data['art'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'codigoObra': codigoObra,
      'art': art,
    };
  }

  @override
  void validate() {
    // TODO: implement validate
  }
}

class Servico extends RequestMapping {
  late final String? itemListaServico;
  late final String? codigoCnae;
  late final String? codigoTributacaoMunicipio;
  late final String? discriminacao;
  late final int? codigoMunicipio;
  late final int? codigoPais;

  ///1: exiExigivel, 2: exiNaoIncidencia, 3: exiIsencao, 4: exiExportaca, 5: exiImunidade, 6: exiSuspensaDecisaoJudicia, 7: exiSuspensaProcessoAdministrativo, 8: exiISSFix
  late final int? exigibilidadeISS;
  late final int? municipioIncidencia;
  late final String? ufPrestacao;

  ///1 - rtTomador, 2 - rtIntermediario
  late final int? responsavelRetencao;
  late final String? operacao;
  late final String? tributacao;
  Servico({
    required this.itemListaServico,
    required this.codigoCnae,
    required this.codigoTributacaoMunicipio,
    this.discriminacao,
    required this.codigoMunicipio,
    this.codigoPais,
    required this.exigibilidadeISS,
    required this.municipioIncidencia,
    required this.ufPrestacao,
    this.responsavelRetencao,
    this.operacao,
    this.tributacao,
  }) : super.empty();
  Servico.empty() : super.empty();
  Servico.fromJson(super.dataRequest) : super.fromJson();
  Servico.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    itemListaServico = data['itemListaServico'];
    codigoCnae = data['codigoCnae'];
    codigoTributacaoMunicipio = data['codigoTributacaoMunicipio'];
    //discriminacao = data['discriminacao'];
    codigoMunicipio = data['codigoMunicipio'];
    codigoPais = data['codigoPais'];
    exigibilidadeISS = data['exigibilidadeISS'];
    municipioIncidencia = data['municipioIncidencia'];
    ufPrestacao = data['ufPrestacao'];
    responsavelRetencao = data['responsavelRetencao'] ?? 1;
    operacao = data['operacao'];
    tributacao = data['tributacao'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'itemListaServico': itemListaServico,
      'codigoCnae': codigoCnae,
      'codigoTributacaoMunicipio': codigoTributacaoMunicipio,
      'discriminacao': discriminacao,
      'codigoMunicipio': codigoMunicipio,
      'codigoPais': codigoPais,
      'exigibilidadeISS': exigibilidadeISS,
      'municipioIncidencia': municipioIncidencia,
      'ufPrestacao': ufPrestacao,
      'responsavelRetencao': responsavelRetencao,
      'operacao': operacao,
      'tributacao': tributacao,
    };
  }

  @override
  void validate() {
    if (itemListaServico == null || itemListaServico!.isEmpty) {
      throw ValidatorException('[itemListaServico] O Item da Lista de Serviço não foi informada');
    }
    if (itemListaServico!.length != 5) throw ValidatorException('[itemListaServico] O Item da Lista de Serviço deve conter 5 carateres Ex: 01.05');

    if (codigoCnae == null || codigoCnae!.isEmpty) throw ValidatorException('[codigoCnae] O Código Cnae do Serviço não foi informado');
    if (codigoCnae!.length != 7) throw ValidatorException('[codigoCnae] O Código Cnae do Serviço deve conter 7 carateres Ex: 6202300');

    if (codigoTributacaoMunicipio == null || codigoTributacaoMunicipio!.isEmpty) {
      throw ValidatorException('[codigoTributacaoMunicipio] O Código da Tributação do Município não foi informado');
    }
    if (codigoTributacaoMunicipio!.length != 5) {
      throw ValidatorException('[codigoTributacaoMunicipio] O Código da Tributação do Município deve conter 5 carateres Ex: 14.02');
    }

    //if (discriminacao == null || discriminacao!.isEmpty) {
    //  throw ValidatorException('[discriminacao] A Discriminação do item do Serviço não foi informado');
    //}

    if (codigoMunicipio == null || codigoMunicipio! == 0) throw ValidatorException('[codigoMunicipio] O Código do Município não foi informado');
    if (codigoMunicipio!.toString().length != 7) {
      throw ValidatorException('[codigoMunicipio] O Código do Município deve conter 7 carateres Ex: 3303807');
    }
    codigoPais ??= 1058;

    if (exigibilidadeISS == null || exigibilidadeISS! == 0) {
      throw ValidatorException('[exigibilidadeISS] O campo da Exigibilidade do ISS não foi informado');
    }
    if (exigibilidadeISS! > 8) {
      throw ValidatorException(
          '[exigibilidadeISS] O campo da Exigibilidade deve conter os valores: 1- exiExigivel, 2- exiNaoIncidencia, 3- exiIsencao, 4- exiExportaca, 5- exiImunidade, 6- exiSuspensaDecisaoJudicia, 7- exiSuspensaProcessoAdministrativo, 8- exiISSFix');
    }
    if (municipioIncidencia == null || municipioIncidencia! == 0) {
      throw ValidatorException('[municipioIncidencia] O Código do Município de Incidencia não foi informado');
    }
    if (municipioIncidencia!.toString().length != 7) {
      throw ValidatorException('[municipioIncidencia] O Código do Município de Incidencia deve conter 7 carateres Ex: 3303807');
    }

    if (ufPrestacao == null || ufPrestacao!.isEmpty) throw ValidatorException('[ufPrestacao] A UF da prestação de serviço não foi informada');
    if (ufPrestacao!.length != 2) throw ValidatorException('[ufPrestacao] A UF da prestação de serviço deve conter 2 carateres Ex: RJ');

    if (responsavelRetencao == null || responsavelRetencao == 0) {
      throw ValidatorException('[responsavelRetencao] A UF da prestação de serviço não foi informada');
    }
    if (responsavelRetencao! > 2) {
      throw ValidatorException(
          '[responsavelRetencao] O campo Responsavel Retenção deve conter os seguintes valores: 1 - rtTomador, 2 - rtIntermediario');
    }
  }
}

class Itens extends RequestMapping {
  late final String? descricao;
  late final String? codServico;
  late final String? codLCServico;
  late final String? itemListaServico;
  late final double? quantidade;
  late final double? valorUnitario;
  late final double? valorDeducoes;
  late final double? valorIss;
  late final double? aliquota;
  late final double? baseCalculo;
  late final double? descontoIncondicionado;
  late final double? valorTotal;
  late final int? tributavel;
  Itens({
    this.descricao,
    this.codServico,
    this.codLCServico,
    this.itemListaServico,
    this.quantidade,
    this.valorUnitario,
    this.valorDeducoes,
    this.valorIss,
    this.aliquota,
    this.baseCalculo,
    this.descontoIncondicionado,
    this.valorTotal,
    this.tributavel,
  }) : super.empty();

  Itens.empty() : super.empty();
  Itens.fromJson(super.dataRequest) : super.fromJson();
  Itens.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    descricao = data['descricao'];
    codServico = data['codServico'];
    codLCServico = data['codLCServico'];
    itemListaServico = data['itemListaServico'];
    quantidade = data['quantidade'];
    valorUnitario = data['valorUnitario'];
    valorDeducoes = data['valorDeducoes'] ?? 0;
    aliquota = data['aliquota'];
    descontoIncondicionado = data['descontoIncondicionado'] ?? 0;
    tributavel = data['tributavel'] ?? 1;

    //calcular campos...
    valorTotal = (valorUnitario! * quantidade!).toDecimal2Us.toDouble();
    baseCalculo = valorTotal! - descontoIncondicionado!;
    valorIss = (baseCalculo! * (aliquota! / 100)).toDecimal2Us.toDouble();
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'descricao': descricao,
      'codServico': codServico,
      'codLCServico': codLCServico,
      'itemListaServico': itemListaServico,
      'quantidade': quantidade,
      'valorUnitario': valorUnitario,
      'valorDeducoes': valorDeducoes,
      'valorIss': valorIss,
      'aliquota': aliquota,
      'baseCalculo': baseCalculo,
      'descontoIncondicionado': descontoIncondicionado,
      'valorTotal': valorTotal,
      'tributavel': tributavel,
    };
  }

  @override
  void validate() {
    if (descricao == null || descricao!.isEmpty) throw ValidatorException('Descrição do item inválida');
    if (quantidade == null || quantidade! == 0) throw ValidatorException('Quantidade do item [$descricao] inválida');
    if (valorUnitario == null || valorUnitario! == 0) throw ValidatorException('Valor unitário do item [$descricao] inválido');
    if (aliquota == null || aliquota! == 0) throw ValidatorException('Alíquota ISS do item [$descricao] inválida');
  }
}

class Valores extends RequestMapping {
  late final double? valorServicos;
  late final double? valorDeducoes;
  late final double? valorPis;
  late final double? aliquotaPis;
  late final double? valorCofins;
  late final double? aliquotaCofins;
  late final double? valorInss;
  late final double? valorIr;
  late final double? valorCsll;
  late final double? iSSRetido;
  late final double? outrasRetencoes;
  late final double? descontoIncondicionado;
  late final double? descontoCondicionado;
  late final double? baseCalculo;
  late final double? aliquota;
  late final double? valorIss;
  late final double? valorIssRetido;
  late final double? valorLiquidoNfse;
  Valores({
    this.valorServicos,
    this.valorDeducoes,
    this.valorPis,
    this.aliquotaPis,
    this.valorCofins,
    this.aliquotaCofins,
    this.valorInss,
    this.valorIr,
    this.valorCsll,
    this.iSSRetido,
    this.outrasRetencoes,
    this.descontoIncondicionado,
    this.descontoCondicionado,
    this.baseCalculo,
    this.aliquota,
    this.valorIss,
    this.valorIssRetido,
    this.valorLiquidoNfse,
  }) : super.empty();
  Valores.empty() : super.empty();
  Valores.fromJson(super.dataRequest) : super.fromJson();
  Valores.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    valorServicos = data['valorServicos'] ?? 0.0;
    valorDeducoes = data['valorDeducoes'] ?? 0.0;
    valorPis = data['valorPis'] ?? 0.0;
    aliquotaPis = data['aliquotaPis'] ?? 0.0;
    valorCofins = data['valorCofins'] ?? 0.0;
    aliquotaCofins = data['aliquotaCofins'] ?? 0.0;
    valorInss = data['valorInss'] ?? 0.0;
    valorIr = data['valorIr'] ?? 0.0;
    valorCsll = data['valorCsll'] ?? 0.0;
    iSSRetido = data['iSSRetido'] ?? 0.0;
    outrasRetencoes = data['outrasRetencoes'] ?? 0.0;
    descontoIncondicionado = data['descontoIncondicionado'] ?? 0.0;
    descontoCondicionado = data['descontoCondicionado'] ?? 0.0;
    baseCalculo = data['baseCalculo'] ?? 0.0;
    aliquota = data['aliquota'] ?? 0.0;
    valorIss = data['valorIss'] ?? 0.0;
    valorIssRetido = data['valorIssRetido'] ?? 0.0;
    valorLiquidoNfse = data['valorLiquidoNfse'] ?? 0.0;
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      if (valorServicos != null) 'valorServicos': valorServicos,
      if (valorDeducoes != null) 'valorDeducoes': valorDeducoes,
      if (valorPis != null) 'valorPis': valorPis,
      if (aliquotaPis != null) 'aliquotaPis': aliquotaPis,
      if (valorCofins != null) 'valorCofins': valorCofins,
      if (aliquotaCofins != null) 'aliquotaCofins': aliquotaCofins,
      if (valorInss != null) 'valorInss': valorInss,
      if (valorIr != null) 'valorIr': valorIr,
      if (valorCsll != null) 'valorCsll': valorCsll,
      if (iSSRetido != null) 'iSSRetido': iSSRetido,
      if (outrasRetencoes != null) 'outrasRetencoes': outrasRetencoes,
      if (descontoIncondicionado != null) 'descontoIncondicionado': descontoIncondicionado,
      if (descontoCondicionado != null) 'descontoCondicionado': descontoCondicionado,
      if (baseCalculo != null) 'baseCalculo': baseCalculo,
      if (aliquota != null) 'aliquota': aliquota,
      if (valorIss != null) 'valorIss': valorIss,
      if (valorIssRetido != null) 'valorIssRetido': valorIssRetido,
      if (valorLiquidoNfse != null) 'valorLiquidoNfse': valorLiquidoNfse,
    };
  }

  @override
  void validate() {
    // TODO: implement validate
  }
}

class CondicaoPagamento extends RequestMapping {
  late final int? qtdParcela;
  late final String? condicao;
  CondicaoPagamento({
    this.qtdParcela,
    this.condicao,
  }) : super.empty();
  CondicaoPagamento.empty() : super.empty();
  CondicaoPagamento.fromJson(super.dataRequest) : super.fromJson();
  CondicaoPagamento.fromMap(super.dataRequest) : super.fromMap();

  @override
  void fromMap() {
    qtdParcela = data['qtdParcela'];
    condicao = data['condicao'];
  }

  @override
  Map<String, dynamic> toMap() {
    return {
      'qtdParcela': qtdParcela,
      'condicao': condicao,
    };
  }

// 'A_VISTA'
// 'NA_APRESENTACAO'
// 'A_PRAZO'
// 'CARTAO_DEBITO'
// 'CARTAO_CREDITO'
  @override
  void validate() {
    // TODO: implement validate
  }
}
